# Import classes, functions, and submodules to be accessible from library
from .constants import Status
from .constants import RequestType
from .constants import RedactMode
from .response import ResponseObject
from .response import cfnresponse
from .response import is_valid_event
from .decorators import decorator
from .decorators import rdecorator
from .decorators import sdecorator
from .redaction import RedactionConfig
from .redaction import StandaloneRedactionConfig
from .redaction import RedactionRuleSet

__all__ = ['Exceptions']



""" Constants for the accustom library.

These are constants are the Custom Resource Request Type Constants
"""


class Status:
    """CloudFormation Custom Resource Status Constants

    https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/crpg-ref-responses.html
    """
    SUCCESS = 'SUCCESS'
    FAILED = 'FAILED'


class RequestType:
    """CloudFormation Custom Resource Request Type Constants

    https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/crpg-ref-requesttypes.html
    """
    CREATE = 'Create'
    UPDATE = 'Update'
    DELETE = 'Delete'


class RedactMode:
    """RedactMode Options"""

    BLACKLIST = 'black'
    WHITELIST = 'white'



""" ResponseObject and cfnresponse function for the accustom library

This allows you to communicate with CloudFormation

"""

# Exceptions
from .Exceptions import DataIsNotDictException
from .Exceptions import NoPhysicalResourceIdException
from .Exceptions import InvalidResponseStatusException
from .Exceptions import FailedToSendResponseException
from .Exceptions import NotValidRequestObjectException

# Constants
from .constants import Status
from .constants import RequestType

# Required Libraries
import json
import logging
import sys
import six
from urllib.parse import urlparse

logger = logging.getLogger(__name__)

# Import Requests
try:
    import requests
except ImportError:
    from botocore.vendored import requests
    logger.warning("botocore.vendored version of requests is deprecated. Please include requests in your code bundle.")

def is_valid_event(event: dict) -> bool:
    """This function takes in a CloudFormation Request Object and checks for the required fields as per:
    https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/crpg-ref-requests.html

    Args:
        event (dict): The request object being processed.

    Returns:
        bool: If the request object is a valid request object

    """
    if not (all(v in event for v in (
        'RequestType',
        'ResponseURL',
        'StackId',
        'RequestId',
        'ResourceType',
        'LogicalResourceId'
    ))):
        # Check we have all the required fields
        return False

    if event['RequestType'] not in [RequestType.CREATE, RequestType.DELETE, RequestType.UPDATE]:
        # Check if the request type is a valid request type
        return False

    scheme = urlparse(event['ResponseURL']).scheme
    if scheme == '' or scheme not in ('http', 'https'):
        # Check if the URL appears to be a valid HTTP or HTTPS URL
        # Technically it should always be an HTTPS URL but hedging bets for testing to allow http
        return False

    if event['RequestType'] in [RequestType.UPDATE, RequestType.DELETE] and 'PhysicalResourceId' not in event:
        # If it is an Update or Delete request there needs to be a PhysicalResourceId key
        return False

    # All checks passed
    return True


def cfnresponse(event: dict, responseStatus: str, responseReason: str = None, responseData: dict = None,
                physicalResourceId: str = None, context: dict = None, squashPrintResponse: bool = False):
    """Format and send CloudFormation Custom Resource Objects

    This section is derived off the cfnresponse source code provided by Amazon:

    https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-lambda-function-code.html

    Creates a JSON payload that is sent back to the ResponseURL (pre-signed S3 URL).

    Args:
        event: A dict containing CloudFormation custom resource request field
        responseStatus (Status.SUCCESS or Status.FAILED): Should have the value of 'SUCCESS' or 'FAILED'
        responseData (dict): The response data to be passed to CloudFormation. If it contains ExceptionThrown
            on FAILED this is given as reason overriding responseReason.
        responseReason (str): The reason for this result.
        physicalResourceId (str): The PhysicalResourceID to be sent back to CloudFormation
        context (context object): Can be used in lieu of a PhysicalResourceId to use the Lambda Context to derive
            an ID.
        squashPrintResponse (boolean): When logging set to debug and this is set to False, it will print the response
            (defaults to False). If set to True this will also send the response with NoEcho set to True.

        Note that either physicalResourceId or context must be defined, and physicalResourceId supersedes
        context

    Returns:
        Dictionary of Response Sent

    Raises:
        NoPhysicalResourceIdException
        InvalidResponseStatusException
        DataIsNotDictException
        FailedToSendResponseException
        NotValidRequestObjectException

    """
    if not is_valid_event(event):
        # If it is not a valid event we need to raise an exception
        message = 'The event object passed is not a valid Request Object as per ' + \
                  'https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/crpg-ref-requests.html'
        logger.error(message)
        raise NotValidRequestObjectException(message)

    if physicalResourceId is None and context is None and 'PhysicalResourceId' not in event:
        raise NoPhysicalResourceIdException("Both physicalResourceId and context are None, and there is no" +
                                            "physicalResourceId in the event")

    if responseStatus != Status.FAILED and responseStatus != Status.SUCCESS:
        raise InvalidResponseStatusException("%s is not a valid status" % responseStatus)

    if responseData is not None and not isinstance(responseData, dict):
        raise DataIsNotDictException("Data provided was not a dictionary")

    if responseStatus == Status.FAILED:
        if responseReason is not None and responseData is not None and 'ExceptionThrown' in responseData:
            responseReason = "There was an exception thrown in execution of '%s'" % responseData['ExceptionThrown']
        elif responseReason is None:
            responseReason = 'Unknown failure occurred'

        if context is not None:
            responseReason = "%s -- See the details in CloudWatch Log Stream: %s" % (responseReason,
                                                                                     context.log_stream_name)

    elif context is not None and responseReason is None:
            responseReason = "See the details in CloudWatch Log Stream: %s" % context.log_stream_name

    responseUrl = event['ResponseURL']

    if physicalResourceId is None and 'PhysicalResourceId' in event: physicalResourceId = event['PhysicalResourceId']

    responseBody = {'Status': responseStatus}
    if responseReason is not None: responseBody['Reason'] = responseReason
    responseBody['PhysicalResourceId'] = physicalResourceId or context.log_stream_name
    responseBody['StackId'] = event['StackId']
    responseBody['RequestId'] = event['RequestId']
    responseBody['LogicalResourceId'] = event['LogicalResourceId']
    if responseData is not None: responseBody['Data'] = responseData 
    if squashPrintResponse: responseBody['NoEcho'] = 'true'

    json_responseBody = json.dumps(responseBody)

    logger.info("Sending response to pre-signed URL.")
    logger.debug("URL: %s" % responseUrl)
    if not squashPrintResponse: logger.debug("Response body:\n" + json_responseBody)

    headers = {
        'content-type': '',
        'content-length': str(len(json_responseBody))
    }

    # Flush the buffers to attempt to prevent log truncations when resource is deleted
    # by stack in next action
    sys.stdout.flush()
    # Flush stdout buffer
    sys.stderr.flush()
    # Flush stderr buffer

    try:
        response = requests.put(responseUrl, data=json_responseBody, headers=headers)
        if response.status_code != 200:
            # Exceptions will only be thrown on timeout or other errors, in order to catch an invalid
            # status code like 403 we will need to explicitly check the status code. In normal operation
            # we should get a "200 OK" response to our PUT.
            message = "Unable to send response to URL, status code received: %d %s" % (response.status_code,
                                                                                       response.reason)
            logger.error(message)
            raise FailedToSendResponseException(message)
        logger.debug("Response status code: %d %s" % (response.status_code, response.reason))

    except FailedToSendResponseException as e:
        # Want to explicitly catch this exception we just raised in order to raise it unmodified
        raise e

    except Exception as e:
        logger.error('Unable to send response to URL, reason given: %s' % str(e))
        raise FailedToSendResponseException(str(e))

    return responseBody


class ResponseObject(object):
    """Class that allows you to init a ResponseObject for easy function writing"""
    def __init__(self, data: dict = None, physicalResourceId: str = None, reason: str = None,
                 responseStatus: str = Status.SUCCESS, squashPrintResponse: bool = False):
        """Init function for the class

        Args:
            data (dict): data to be passed in the response. Must be a dict if used
            physicalResourceId (str): Physical resource ID to be used in the response
            reason (str): Reason to pass back to CloudFormation in the response Object
            responseStatus (Status.SUCCESS or Status.FAILED): response Status to use in the response Object,
                defaults to SUCCESS
            squashPrintResponse (boolean): When logging set to debug and this is set to False, it will print the
                response (defaults to False). If set to True it will also be sent with NoEcho set to true.

        Raises:
            DataIsNotDictException
            TypeError

        """
        if data is not None and not isinstance(data, dict):
            raise DataIsNotDictException("Data provided was not a dictionary")

        if not isinstance(physicalResourceId, six.string_types) and physicalResourceId is not None:
            raise TypeError('physicalResourceId must be of type string')

        if not isinstance(reason, six.string_types) and reason is not None:
            raise TypeError('message must be of type string')

        if responseStatus != Status.SUCCESS and responseStatus != Status.FAILED:
            raise TypeError('Invalid response status')

        if not isinstance(squashPrintResponse, bool):
            raise TypeError('squashPrintResponse must be of boolean type')

        self.data = data
        self.physicalResourceId = physicalResourceId
        self.reason = reason
        self.responseStatus = responseStatus
        self.squashPrintResponse = squashPrintResponse

    def send(self, event: dict, context: dict = None):
        """Send this CloudFormation Custom Resource Object

        Creates a JSON payload that is sent back to the ResponseURL (pre-signed S3 URL) based upon this response object
        using the cfnresponse function.

        Args:
            event: A dict containing CloudFormation custom resource request field
            context: Can be used in lieu of a PhysicalResourceId to use the Lambda Context to derive an ID.

            Note that either physicalResourceId in the object or context must be defined, and physicalResourceId
            supersedes context

        Returns:
            Dictionary of Response Sent

        Raises:
            NoPhysicalResourceIdException
            InvalidResponseStatusException
            DataIsNotDictException
            FailedToSendResponseException
            NotValidRequestObjectException
        """
        return cfnresponse(event, self.responseStatus, self.reason, self.data, self.physicalResourceId, context,
                           self.squashPrintResponse)

    def __str__(self):
        return 'Response(Status=%s)' % self.responseStatus

    def __repr__(self):
        return str(self)



""" Exceptions for the accustom library.

These are the exceptions that can be returned by accustom
"""


class CannotApplyRuleToStandaloneRedactionConfig(Exception):
    """Indicates that a second rule set was attempted to be applied to a standalone"""
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ConflictingValue(Exception):
    """Indicates that there is already a record with this value"""
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class NoPhysicalResourceIdException(Exception):
    """Indicates that there was no valid value to use for PhysicalResourceId"""
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class InvalidResponseStatusException(Exception):
    """Indicates that there response code was not SUCCESS or FAILED"""
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class DataIsNotDictException(Exception):
    """Indicates that a Dictionary was not passed as Data"""
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class FailedToSendResponseException(Exception):
    """Indicates there was a problem sending the response"""
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)

class NotValidRequestObjectException(Exception):
    """Indicates that the event passed in is not a valid Request Object"""
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)



# Bring exceptions into the root of the submodule

from .exceptions import CannotApplyRuleToStandaloneRedactionConfig
from .exceptions import ConflictingValue
from .exceptions import NoPhysicalResourceIdException
from .exceptions import InvalidResponseStatusException
from .exceptions import DataIsNotDictException
from .exceptions import FailedToSendResponseException
from .exceptions import NotValidRequestObjectException


""" RedactionConfig and StandaloneRedactConfig for the accustom library

This allows you to define a redaction policy for accustom
"""

from .response import is_valid_event
from .constants import RedactMode
from .Exceptions import ConflictingValue
from .Exceptions import CannotApplyRuleToStandaloneRedactionConfig
from .Exceptions import NotValidRequestObjectException

import logging
import six
import re
import copy

logger = logging.getLogger(__name__)

_RESOURCEREGEX_DEFAULT = '^.*$'
REDACTED_STRING = '[REDACTED]'


class RedactionRuleSet(object):
    """Class that allows you to define a redaction rule set for accustom"""

    def __init__(self, resourceRegex: str = _RESOURCEREGEX_DEFAULT):
        """Init function for the class

        Args:
            resourceRegex (String): The regex used to work out what resources to apply this too.

        Raises:
            TypeError

        """
        if not isinstance(resourceRegex, six.string_types):
            raise TypeError('resourceRegex must be a string')

        self.resourceRegex = resourceRegex
        self._properties = []

    def add_property_regex(self, propertiesRegex : str):
        """Allows you to add a property regex to whitelist/blacklist

        Args:
            propertiesRegex (String): The regex used to work out what properties to whitelist/blacklist

        Raises:
            TypeError

        """
        if not isinstance(propertiesRegex, six.string_types):
            raise TypeError('propertiesRegex must be a string')
        self._properties.append(propertiesRegex)

    def add_property(self, propertyName: str):
        """Allows you to add a specific property to whitelist/blacklist

        Args:
            propertyName (String): The name of the property to whitelist/blacklist

        Raises:
            TypeError

        """
        if not isinstance(propertyName, six.string_types):
            raise TypeError('propertyName must be a string')
        self._properties.append('^' + propertyName + '$')


class RedactionConfig(object):
    """Class that allows you define a redaction policy for accustom"""

    def __init__(self, redactMode: str = RedactMode.BLACKLIST, redactResponseURL: bool = False):
        """Init function for the class

        Args:
            redactMode (RedactMode.BLACKLIST or RedactMode.WHITELIST): Determine if we should whitelist or blacklist
            resources, defaults to blacklist
        redactResponseURL (boolean): Prevents the pre-signed URL from being printed preventing out of band responses
        (recommended for production)

        Raises:
            TypeError

        """
        if redactMode != RedactMode.BLACKLIST and redactMode != RedactMode.WHITELIST:
            raise TypeError('Invalid Redaction Type')

        if not isinstance(redactResponseURL, bool):
            raise TypeError('redactResponseURL must be of boolean type')

        self.redactMode = redactMode
        self.redactResponseURL = redactResponseURL
        self._redactProperties = {}

    def add_rule_set(self, ruleSet: RedactionRuleSet):
        """ This function will add a RedactionRuleSet object to the RedactionConfig.

        Args:
            ruleSet (RedactionRuleSet): The rule to be added to the RedactionConfig

        Raises:
            TypeError
            ConflictingValue

        """

        if not isinstance(ruleSet, RedactionRuleSet):
            raise TypeError('Please use RedactionRuleSet class')
        if ruleSet.resourceRegex in self._redactProperties:
            raise ConflictingValue('There is already a record set for resource of regex: %s' % ruleSet.resourceRegex)

        self._redactProperties[ruleSet.resourceRegex] = ruleSet._properties

    def _redact(self, event: dict):
        """ Internal Function. Not to be consumed outside of accustom Library.

            This function will take in an event and return the event redacted as per the redaction config.
        """
        if not is_valid_event(event):
            # If it is not a valid event we need to raise an exception
            message = 'The event object passed is not a valid Request Object as per ' + \
                      'https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/crpg-ref-requests.html'
            logger.error(message)
            raise NotValidRequestObjectException(message)
        ec = copy.deepcopy(event)
        if self.redactMode == RedactMode.WHITELIST:
            if 'ResourceProperties' in ec: ec['ResourceProperties'] = {}
            if 'OldResourceProperties' in ec: ec['OldResourceProperties'] = {}
        for resourceRegex, propertyRegex in self._redactProperties.items():
            if re.search(resourceRegex, event['ResourceType']) is not None:
                # Go through the Properties looking to see if they're in the ResourceProperties or OldResourceProperties
                for index, item in enumerate(propertyRegex):
                    r = re.compile(item)
                    if self.redactMode == RedactMode.BLACKLIST:
                        if 'ResourceProperties' in ec:
                            for mitem in filter(r.match, ec['ResourceProperties']):
                                ec['ResourceProperties'][mitem] = REDACTED_STRING
                        if 'OldResourceProperties' in ec:
                            for mitem in filter(r.match, ec['OldResourceProperties']):
                                ec['OldResourceProperties'][mitem] = REDACTED_STRING
                    elif self.redactMode == RedactMode.WHITELIST:
                        if 'ResourceProperties' in ec:
                            for mitem in filter(r.match, event['ResourceProperties']):
                                ec['ResourceProperties'][mitem] = event['ResourceProperties'][mitem]
                        if 'OldResourceProperties' in ec:
                            for mitem in filter(r.match, event['OldResourceProperties']):
                                ec['OldResourceProperties'][mitem] = event['OldResourceProperties'][mitem]
        if self.redactMode == RedactMode.WHITELIST:
            if 'ResourceProperties' in ec:
                for key, value in event['ResourceProperties'].items():
                    if key not in ec['ResourceProperties']: ec['ResourceProperties'][key] = REDACTED_STRING
            if 'OldResourceProperties' in ec:
                for key, value in event['OldResourceProperties'].items():
                    if key not in ec['OldResourceProperties']: ec['OldResourceProperties'][key] = REDACTED_STRING

        if self.redactResponseURL: del ec['ResponseURL']
        return ec

    def __str__(self):
        return 'RedactionConfig(%s)' % self.redactMode

    def __repr__(self):
        return str(self)


class StandaloneRedactionConfig(RedactionConfig):
    def __init__(self, ruleSet: RedactionRuleSet, redactMode: str = RedactMode.BLACKLIST,
                 redactResponseURL: bool = False):
        """Init function for the class

        Args:
            redactMode (RedactMode.BLACKLIST or RedactMode.WHITELIST): Determine if we should whitelist or blacklist
            resources, defaults to blacklist
        redactResponseURL (boolean): Prevents the pre-signed URL from being printed preventing out of band responses
        (recommended for production)
        ruleSet (RedactionRuleSet): The single rule set to be applied

        Raises:
            TypeError
        """

        RedactionConfig.__init__(self, redactMode=redactMode, redactResponseURL=redactResponseURL)
        ruleSet.resourceRegex = _RESOURCEREGEX_DEFAULT
        # override resource regex to be default
        assert (ruleSet is not None)
        RedactionConfig.add_rule_set(self, ruleSet)

    def add_rule_set(self, ruleSet: RedactionRuleSet):
        """ Overrides the add_rule_set operation with one that will immediately throw an exception

            Raises
                CannotApplyRuleToStandaloneRedactionConfig
        """
        raise CannotApplyRuleToStandaloneRedactionConfig('StandaloneRedactionConfig does not support additional rules.')



""" Decorators for the accustom library.

This includes two decorators, one for the handler function, and one to apply to any resource handling
functions.
"""

# Exceptions
from .Exceptions import FailedToSendResponseException
from .Exceptions import DataIsNotDictException
from .Exceptions import InvalidResponseStatusException
from .Exceptions import NotValidRequestObjectException

# Constants
from .constants import RequestType
from .constants import Status

# Response
from .response import is_valid_event

# RedactionConfig
from .redaction import RedactionConfig
from .redaction import StandaloneRedactionConfig

# Required Libraries
import logging
import json
from functools import wraps
from uuid import uuid4
from .response import ResponseObject
import six
from boto3 import client
from botocore.client import Config
from botocore import exceptions as bexceptions

logger = logging.getLogger(__name__)

# Import Requests
try:
    import requests
except ImportError:
    from botocore.vendored import requests
    logger.warning("botocore.vendored version of requests is deprecated. Please include requests in your code bundle.")
import urllib3.exceptions


# Time in milliseconds to set the alarm for (in milliseconds)
# Should be set to twice the worst case response time to send to S3
# Setting to 2 seconds for safety
TIMEOUT_THRESHOLD = 2000


def decorator(enforceUseOfClass: bool = False, hideResourceDeleteFailure: bool = False,
              redactConfig: RedactionConfig = None, timeoutFunction: bool = False):
    """Decorate a function to add exception handling and emit CloudFormation responses.

    Usage with Lambda:
        import accustom
        @accustom.decorator()
        def function_handler(event, context)
            sum = (float(event['ResourceProperties']['key1']) +
                   float(event['ResourceProperties']['key2']))
            return { 'sum' : sum }

    Usage outside Lambda:
        import accustom
        @accustom.decorator()
        def function_handler(event)
            sum = (float(event['ResourceProperties']['key1']) +
                   float(event['ResourceProperties']['key2']))
            r = accustom.ResponseObject(data={'sum':sum},physicalResourceId='abc')
            return r

    Args:
        enforceUseOfClass (boolean): When true send a FAILED signal if a ResponseObject class is not utilised.
            This is implicitly set to true if no Lambda Context is provided.
        hideResourceDeleteFailure (boolean): When true will return SUCCESS even on getting an Exception for
            DELETE requests.
        redactConfig (RedactionConfig): Configuration of how to redact the event object.
        timeoutFunction (boolean): Will automatically send a failure signal to CloudFormation before Lambda timeout
            provided that this function is executed in Lambda

    Returns:
        dict: The response object sent to CloudFormation

    Raises:
        FailedToSendResponseException
        NotValidRequestObjectException

    Decorated Function Arguments:
        event (dict): The request object being processed (Required).
        context (dict): The Lambda context of this execution (optional)

    """

    def inner_decorator(func):
        @wraps(func)
        def handler_wrapper(event: dict, context: dict = None):
            nonlocal redactConfig
            nonlocal timeoutFunction
            logger.info('Request received, processing...')
            if not is_valid_event(event):
                # If it is not a valid event we need to raise an exception
                message = 'The event object passed is not a valid Request Object as per ' + \
                          'https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/crpg-ref-requests.html'
                logger.error(message)
                raise NotValidRequestObjectException(message)

            # Timeout Function Handler
            if 'LambdaParentRequestId' in event:
                logger.info('This request has been invoked as a child, for parent logs please see request ID: %s' %
                            event['LambdaParentRequestId'])
            elif context is None and timeoutFunction:
                logger.warning('You cannot use the timeoutFunction option outside of Lambda. To suppress this warning' +
                               ', set timeoutFunction to False')
            elif timeoutFunction:
                # Attempt to invoke the function. Depending on the error we get may continue execution or return
                logger.info('Request has been invoked in Lambda with timeoutFunction set, attempting to invoke self')
                pevent = event.copy()
                pevent['LambdaParentRequestId'] = context.aws_request_id
                payload = json.dumps(pevent).encode('UTF-8')
                timeout = (context.get_remaining_time_in_millis() - TIMEOUT_THRESHOLD) / 1000
                # Edge case where time is set to very low timeout, use half the timeout threshold as the timeout for the
                # the Lambda Function
                if timeout <= 0: timeout = TIMEOUT_THRESHOLD / 2000
                config = Config(connect_timeout=2, read_timeout=timeout, retries={'max_attempts': 0})
                b_lambda = client('lambda', config=config)

                # Normally we would just do a catch all error handler but in this case we want to be paranoid
                try:
                    response = b_lambda.invoke(FunctionName=context.invoked_function_arn,
                                               InvocationType='RequestResponse', Payload=payload)
                    # Further checks
                    if 'FunctionError' in response:
                        response.get('Payload', ''.encode('UTF-8'))
                        message = 'Invocation got an error: %s' % payload.decode()
                        logger.error(message)
                        return ResponseObject(reason=message, responseStatus=Status.FAILED).send(event, context)
                    else:
                        # In this case the function returned without error which means we can assume the chained
                        # invokation sent a response, so we do not have too.
                        logger.info('Compeleted execution of chained invocation, returning payload')
                        response.get('Payload', ''.encode('UTF-8'))
                        return payload.decode()

                except bexceptions.ClientError as e:
                    logger.warning('Caught exception %s while trying to invoke function. Running handler locally.'
                                   % str(e))
                    logger.warning('You cannot use the timeoutFunction option without the ability for the function to' +
                                   ' invoke itself. To suppress this warning, set timeoutFunction to False')
                except bexceptions.ConnectionError as e:
                    logger.error('Got error %s while trying to invoke function. Running handler locally' % str(e))
                    logger.error('You cannot use the timeoutFunction option without the ability to connect to the ' +
                                 'Lambda API from within the function. As we may not have time to execute the ' +
                                 'function, returning failure.')
                    return ResponseObject(reason='Unable to call Lambda to do chained invoke, returning failure.',
                                          responseStatus=Status.FAILED).send(event,context)
                except bexceptions.ReadTimeoutError:
                    # This should be a critical failure
                    logger.error('Waited the read timeout and function did not return, returning an error')
                    return ResponseObject(reason='Lambda function timed out, returning failure.',
                                          responseStatus=Status.FAILED).send(event, context)
                except Exception as e:
                    message = 'Got an %s I did not understand while trying to invoke child function: %s' % (e.__class__,
                                                                                                            str(e))
                    logger.error(message)
                    return ResponseObject(reason=message, responseStatus=Status.FAILED).send(event, context)

            # Debug Logging Handler
            if logger.getEffectiveLevel() <= logging.DEBUG:
                if context is not None:
                    logger.debug('Running request with Lambda RequestId: %s' % context.aws_request_id)
                if redactConfig is not None and isinstance(redactConfig, (StandaloneRedactionConfig, RedactionConfig)):
                    logger.debug('Request Body:\n' + json.dumps(redactConfig._redact(event)))
                elif redactConfig is not None:
                    logger.warning('A non valid RedactionConfig was provided, and ignored')
                    logger.debug('Request Body:\n' + json.dumps(event))
                else:
                    logger.debug('Request Body:\n' + json.dumps(event))

            try:
                logger.info('Running CloudFormation request %s for stack: %s' % (event['RequestId'], event['StackId']))
                # Run the function
                if context is not None:
                    result = func(event, context)
                else:
                    result = func(event)

            except Exception as e:
                # If there was an exception thrown by the function, send a failure response
                result = ResponseObject(
                    physicalResourceId=str(uuid4()) if context is None else None,
                    reason='Function %s failed due to exception "%s"' % (func.__name__, str(e)),
                    responseStatus=Status.FAILED)
                logger.error(result.reason)

            if not isinstance(result, ResponseObject):
                # If a ResponseObject is not provided, work out what kind of response object to pass, or return a
                # failure if it is an invalid response type, or if the enforceUseOfClass is explicitly or implicitly set
                if context is None:
                    result = ResponseObject(
                        reason='Response Object of type %s was not a ResponseObject and there is no Lambda Context'
                               % result.__class__,
                        responseStatus=Status.FAILED)
                    logger.error(result.reason)
                elif enforceUseOfClass:
                    result = ResponseObject(
                        reason='Response Object of type %s was not a ResponseObject instance and ' +
                               'enforceUseOfClass set to true' % result.__class__,
                        responseStatus=Status.FAILED)
                    logger.error(result.reason)
                elif result is False:
                    result = ResponseObject(
                        reason='Function %s returned False.' % func.__name__,
                        responseStatus=Status.FAILED)
                    logger.debug(result.reason)
                elif isinstance(result, dict):
                    result = ResponseObject(data=result)
                elif isinstance(result, six.string_types):
                    result = ResponseObject(data={'Return': result})
                elif result is None or result is True:
                    result = ResponseObject()
                else:
                    result = ResponseObject(
                        reason='Return value from Function %s is of unsupported type %s' % (func.__name__,
                                                                                            result.__class__),
                        responseStatus=Status.FAILED)
                    logger.error(result.reason)

            # This block will hide resources on delete failure if the flag is set to true
            if event['RequestType'] == RequestType.DELETE and result.responseStatus == Status.FAILED \
                    and hideResourceDeleteFailure:
                logger.warning('Hiding Resource DELETE request failure')
                if result.data is not None:
                    if not result.squashPrintResponse:
                        logger.debug('Data:\n' + json.dumps(result.data))
                    else:
                        logger.debug('Data: [REDACTED]')
                if result.reason is not None: logger.debug('Reason: %s' % result.reason)
                if result.physicalResourceId is not None: logger.debug('PhysicalResourceId: %s'
                                                                       % result.physicalResourceId)
                result = ResponseObject(
                    reason='There may be resources created by this Custom Resource that have not been cleaned' +
                           'up despite the fact this resource is in DELETE_COMPLETE',
                    physicalResourceId=result.physicalResourceId,
                    responseStatus=Status.SUCCESS)

            try:
                return_value = result.send(event, context)
            except Exception as e:
                if isinstance(e, FailedToSendResponseException):
                    raise e
                logger.error('Malformed request, Exception: %s' % str(e))
                if result.data is not None and not isinstance(e, DataIsNotDictException):
                    if not result.squashPrintResponse:
                        logger.debug('Data:\n' + json.dumps(result.data))
                    else:
                        logger.debug('Data: [REDACTED]')
                if result.reason is not None: logger.debug('Reason: %s' % result.reason)
                if result.physicalResourceId is not None: logger.debug('PhysicalResourceId: %s'
                                                                       % result.physicalResourceId)
                if not isinstance(e, InvalidResponseStatusException): logger.debug('Status: %s'
                                                                                   % result.responseStatus)
                result = ResponseObject(
                    reason='Malformed request, Exception: %s' % str(e),
                    physicalResourceId=result.physicalResourceId,
                    responseStatus=Status.FAILED)
                return_value = result.send(event, context)
            return return_value

        return handler_wrapper

    return inner_decorator


def rdecorator(decoratorHandleDelete: bool = False, expectedProperties: list = None, genUUID: bool = True):
    """Decorate a function to add input validation for resource handler functions.

        Usage with Lambda:
            import accustom
            @accustom.rdecorator(expectedProperties=['key1','key2'],genUUID=False)
            def resource_function(event, context):
                sum = (float(event['ResourceProperties']['key1']) +
                       float(event['ResourceProperties']['key2']))
                return { 'sum' : sum }
            @accustom.decorator()
            def function_handler(event, context)
                return resource_function(event,context)

        Usage outside Lambda:
            import accustom
            @accustom.rdecorator(expectedProperties=['key1','key2'])
            def resource_function(event, context=None)
                sum = (float(event['ResourceProperties']['key1']) +
                       float(event['ResourceProperties']['key2']))
                r = accustom.ResponseObject(data={'sum':sum},physicalResourceId=event['PhysicalResourceId'])
                return r
            @accustom.decorator()
            def function_handler(event)
                return resource_function(event)

        Args:
            decoratorHandleDelete (boolean): When set to true, if a delete request is made in event the decorator will
                return a ResponseObject with a with SUCCESS without actually executing the decorated function
            genUUID (boolean): When set to true, if the PhysicalResourceId in the event is not set, automatically
                generate a UUID4 and put it in the PhysicalResourceId field.
            expectedProperties (list of expected properties): Pass in a list or tuple of properties that you want to
                check for before running the decorated function.

        Returns:
            The result of the decorated function, or a ResponseObject with SUCCESS depending on the event and flags.

        Raises:
            NotValidRequestObjectException
            Any exception raised by the decorated function.


        Decorated Function Arguments:
            event (dict): The request object being processed (Required).
    """

    def resource_decorator_inner(func):
        @wraps(func)
        def resource_decorator_handler(event: dict, *args, **kwargs):
            if not is_valid_event(event):
                # If it is not a valid event we need to raise an exception
                message = 'The event object passed is not a valid Request Object as per ' + \
                          'https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/crpg-ref-requests.html'
                logger.error(message)
                raise NotValidRequestObjectException(message)
            logger.info('Supported resource %s' % event['ResourceType'])

            # Set the Physical Resource ID to a randomly generated UUID if it is not present
            if genUUID and 'PhysicalResourceId' not in event:
                event['PhysicalResourceId'] = str(uuid4())
                logger.info('Set PhysicalResourceId to %s' % event['PhysicalResourceId'])

            # Handle Delete when decoratorHandleDelete is set to True
            if decoratorHandleDelete and event['RequestType'] == RequestType.DELETE:
                logger.info('Request type %s detected, returning success without calling function' % RequestType.DELETE)
                return ResponseObject(physicalResourceId=event['PhysicalResourceId'])

            # Validate important properties exist
            if expectedProperties is not None and isinstance(expectedProperties, (list, tuple)):
                for index, item in enumerate(expectedProperties):
                    if item not in event['ResourceProperties']:
                        err_msg = 'Property %s missing, sending failure signal' % item
                        logger.info(err_msg)
                        return ResponseObject(reason=err_msg, responseStatus=Status.FAILED,
                                              physicalResourceId=event['PhysicalResourceId'])

            # If a list or tuple was not provided then log a warning
            elif expectedProperties is not None:
                logger.warning('expectedProperties passed to decorator is not a list, properties were not validated.')

            # Pre-validation complete, calling function
            return func(event, *args, **kwargs)

        return resource_decorator_handler

    return resource_decorator_inner


def sdecorator(decoratorHandleDelete: bool = False, expectedProperties: list = None, genUUID: bool = True,
               enforceUseOfClass: bool = False, hideResourceDeleteFailure: bool = False,
               redactConfig: RedactionConfig = None, timeoutFunction: bool = True):
    """Decorate a function to add input validation for resource handler functions, exception handling and send
    CloudFormation responses.


    Usage with Lambda:
        import accustom
        @accustom.sdecorator(expectedProperties=['key1','key2'],genUUID=False)
        def resource_handler(event, context):
            sum = (float(event['ResourceProperties']['key1']) +
                   float(event['ResourceProperties']['key2']))
            return { 'sum' : sum }

    Usage outside Lambda:
        import accustom
        @accustom.sdecorator(expectedProperties=['key1','key2'])
        def resource_handler(event, context=None)
            sum = (float(event['ResourceProperties']['key1']) +
                   float(event['ResourceProperties']['key2']))
            r = accustom.ResponseObject(data={'sum':sum},physicalResourceId=event['PhysicalResourceId'])
            return r

    Args:
        decoratorHandleDelete (boolean): When set to true, if a delete request is made in event the decorator will
            return SUCCESS to CloudFormation without actually executing the decorated function
        genUUID (boolean): When set to true, if the PhysicalResourceId in the event is not set, automatically generate
            a UUID4 and put it in the PhysicalResourceId field.
        expectedProperties (list of expected properties): Pass in a list or tuple of properties that you want to check
            for before running the decorated function.
        enforceUseOfClass (boolean): When true send a FAILED signal if a ResponseObject class is not utilised.
            This is implicitly set to true if no Lambda Context is provided.
        hideResourceDeleteFailure (boolean): When true will return SUCCESS even on getting an Exception for DELETE
            requests. Note that this particular flag is made redundant if decoratorHandleDelete is set to True.
        redactConfig (StandaloneRedactionConfig): Configuration of how to redact the event object.
        timeoutFunction (boolean): Will automatically send a failure signal to CloudFormation 1 second before Lambda
            timeout provided that this function is executed in Lambda

    Returns:
        The response object sent to CloudFormation

    Raises:
         FailedToSendResponseException
         NotValidRequestObjectException
    """
    if not isinstance(redactConfig, StandaloneRedactionConfig) and logger.getEffectiveLevel() <= logging.DEBUG:
        logger.warning('A non valid StandaloneRedactionConfig was provided, and ignored')
        redactConfig = None

    def standalone_decorator_inner(func):
        @wraps(func)
        @decorator(enforceUseOfClass=enforceUseOfClass, hideResourceDeleteFailure=hideResourceDeleteFailure,
                   redactConfig=redactConfig, timeoutFunction=timeoutFunction)
        @rdecorator(decoratorHandleDelete=decoratorHandleDelete, expectedProperties=expectedProperties, genUUID=genUUID)
        def standalone_decorator_handler(event: dict, context: dict = None):
            return func(event, context)

        return standalone_decorator_handler

    return standalone_decorator_inner



from distutils.core import setup
from setuptools import find_packages

setup(
    name = 'accustom',
    version = '1.1.0',
    description = 'Custom resource library for AWS CloudFormation',
    long_description = 'Accustom is a library for responding to Custom Resources in AWS CloudFormation using the decorator pattern.',
    url = 'https://github.com/NightKhaos/accustom',
    author = 'Taylor Bertie',
    author_email = 'nightkhaos@gmail.com',
    license = 'MIT',
    download_url = 'https://github.com/NightKhaos/accustom/archive/1.1.0.tar.gz',
    keywords = ['cloudformation','lambda','custom','resource','decorator'],
    classifiers = [
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Application Frameworks',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
    packages = ['accustom', 'accustom.Exceptions'],
    install_requires=[
        'botocore>=1.10',
        'boto3>=1.8',
        'requests>=2.0'
    ],
    python_requires='>=3, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*, <4',
    )
